#
# link.awk : generate Linker Script File: link.cmd
#   by uratan! 2014.10.17
#
	#
	# LPC810 memories
	#
	# Boot-ROM
	#   0x1FFF_0000--0x1FFF_1FFF (8kB)
	#
	# on-chip-flash
	#   0x0000_0000--0x0000_0FFF (4kB)   (16kB @ LPC812)
	#
	# RAM
	#   0x1000_0000--0x1000_03FF (1kB)   (4kB @ LPC812)
	#  (0x1000_0400--0x1000_0FFF (3kB)   (image of 1kB ram @ LPC810))
	#  (0x1400_0000--0x1400_03FF (1KB) MTB registers)
	#
	# ???
	#   0x0c00_0000--0x0c00_01FF (512B)
	#
BEGIN {
	#
	# options
	#
	copy_and_execute_on_ram =     0;	# enable codes-to-copy in Startup.S !
	has_sections_for_romexec =    0;	# (to place specified funcs on ROM)
	has_sections_for_spifi_lib =  0;	# (spifi_drv_M3.lib)
	has_sections_for_arm_newlib = 0;	# (sometimes newlib attached)

	####################################### script start
	if(copy_and_execute_on_ram) {
		Rom_Ram = "RAM";
	} else {
		Rom_Ram = "ROM";
	}
	pr("/*");
	pr(" * link.cmd : Linker Script File");
	pr(" *   see link.awk for comments and structure");
	pr(" */");
	####################################### Memory Definitions
	pr("MEMORY");
	pr("{");
	pr("	ROM (rx) : ORIGIN = 0x00000000, LENGTH = 4k	/* flash */");
	pr("	RAM (rwx): ORIGIN = 0x10000000, LENGTH = 1k	/* SRAM */");
	pr("}");
	####################################### Section Definitions
	### syntax:
	###   output-section-xxx : { input-file(input-section-yyy) } > VMA-location
	###   output-section-xxx : { inp...                        } > ...
	###
	pr("SECTIONS");
	pr("{");
	####################################### first output section is .boot, executed directly from flash
	pr("	.boot :");
	pr("	{");
	pr("		KEEP(*(.vectorg))");
	pr("		. = ALIGN(16);");
	pr("		KEEP(*(.init))");		### place startup codes on here
	pr("		. = ALIGN(16);");
if(has_sections_for_romexec) {
	pr("		KEEP(*(.romexec))");	### to pin specified funcs on ROM
	pr("		. = ALIGN(16);");
}
	pr("	} > ROM");					### <====== always on ROM
	pr("	_boot_end = . ;");
	####################################### .exec output section is for execute codes
if(copy_and_execute_on_ram) {
	pr("	.exec : AT(_boot_end)");
} else {
	pr("	.exec :");
}
	pr("	{");
	pr("		_exec_start = . ;");
	pr("		*(.text)");				### executives (default)
	pr("		*(.text.*)");
	pr("		. = ALIGN(16);");
	pr("		*(.rodata)");			### read-only data (constants)
	pr("		*(.rodata.*)");
	pr("		. = ALIGN(16);");
	####################################### for spifi_drv_M3.lib
if(has_sections_for_spifi_lib) {
	pr("		*(.rev16_text)");
	pr("		*(.revsh_text)");
	pr("		*(.constdata)");
	pr("		. = ALIGN(16);");
}
	pr("	} > " Rom_Ram);				### <====== switchable
	####################################### for ARM newlib specials
if(has_sections_for_arm_newlib) {
	pr("	.exec_extab :");
	pr("	{");
	pr("		*(.ARM.extab*)");
	pr("		. = ALIGN(16);");
	pr("	} > " Rom_Ram);				### <====== switchable
	pr("	__exidx_start = .;");
	pr("	.exec_exidx :");
	pr("	{");
	pr("		*(.ARM.exidx*)");
	pr("	} > " Rom_Ram);				### <====== switchable
	pr("	__exidx_end = .;");
}
	####################################### .exec conclusion
	pr("	_exec_end = . ;");
	pr("	_lma_exec_start = LOADADDR(.exec);");
	####################################### .data output section is for initialized data
if(copy_and_execute_on_ram) {
	pr("	.data :");
} else {
	pr("	.data : AT(_exec_end)");
}
	pr("	{");
	pr("		_data_start = .;");
	pr("		*(.data)");
	pr("		*(.data.*)");
	pr("		. = ALIGN(16);");
	pr("	} > RAM");					### <====== always on RAM
	pr("	_data_end = . ;");
	pr("	_lma_data_start = LOADADDR(.data);");
	####################################### .bss output section is for uninitialized data
	pr("	.bss (NOLOAD) :");
	pr("	{");
	pr("		_bss_start = . ;");
	pr("		*(.bss)");
	pr("		*(COMMON)");
	pr("		. = ALIGN(16);");
	pr("	} > RAM");					### <====== always on RAM
	pr("	_bss_end = . ;");
	####################################### .stack output section is simply for stack
	pr("	.stack ALIGN(16) :");
	pr("	{");
	pr("		*(.stack)");
	pr("	} > RAM");					### <====== always on RAM
	#######################################
	pr("}");
}

#
# print string with RETURN
#
function pr(s)
{
	print s;
}
