/*
 * main.c - main for Ir-Transceiver
 *   by uratan! 2014.10.16
 */
//#define __MAIN_C__		/* <== inform common.h to alloc variables */
#include "common.h"
#include <chip.h>
	/*
	 * my LPC810 (DIP8)
	 *   +------------+
	 *   | LPC810     |
	 *   ) P  M021F   |
	 *   ) X  1M28603 |
	 *   | N  +D3414C | ... year 3, week 41, '4C' ver.13.4
	 *   +------------+
	 *      | Dev.ID:  0x00008100
	 *      | B.L.Ver: 13.4
	 */
/*
 * I/O pin usage
 *   pin#5: PIO0_1 (input)  - push-SW
 *   pin#3: PIO0_3 (output) - 38kHz carrier by SCT
 *   pin#4: PIO0_2 (output) - U1_TxD for Ir LED
 */
#define GPI_pSW    (1<<1)
#if EXP_MA_T33
#define GPO_IR     (1<<2)
#endif /* EXP_MA_T33 */

/*
 * use external clock thru CLKIN
 */
#define USE_CLKIN_12MHz  0

/*
 * SysTick base interval
 */
#define SYSTICK_Hz   100		/* get 10msec interval */

#if EXP_MA_T33
/*
 * slow clock is enough
 */
#define DIV  4
#define myLOOP_WAIT_1msec   (864/DIV)		/* 1.0msec @ 12MHz,  1-wait, -Os */
#endif /* EXP_MA_T33 */

/*
 * functions
 */
#if EXP_MA_T33
void make_long(void);
void make_p2(void);
void make_p1(void);
void make_break(void);
void make_function(char const *f);
void make_request(char c);
#endif /* EXP_MA_T33 */
int main(void);
void configure_fixed_pins(void);
u_long initial_sysCLK(int use_PLL_or_set_divider);
void configure_GPIO(u_long mainCLK);
void init_SysTick(u_long sysCLK);
void SysTick_Handler(void)        ATTR_INTR;
#if POWER_DOWN
void PININT0_Handler(void)        ATTR_INTR;
#endif /* POWER_DOWN */
void def_Exception_Handler(void)  ATTR_INTR;
void def_NMI_Handler(void)        ATTR_INTR;
void def_HardFault_Handler(void)  ATTR_INTR;
void def_SVCall_Handler(void)     ATTR_INTR;
void def_PendSV_Handler(void)     ATTR_INTR;

/*
 * variables
 */
VOLATILE int systick_raw;

/*
 * externs
 */
#if EXP_MA_T33
	/* no more use uart1 */
#else /* EXP_MA_T33 */
/* in uart1.c */
extern void uart1_c_init(u_long baudrate, u_long mainCLK);
extern void u1_Putchar(int ch);
#endif /* EXP_MA_T33 */
/* in sctimer.c */
extern void sctimer_c_init(u_long sysCLK);
extern void sct_enable_carrier(void);
extern void sct_disable_carrier(void);


/* =============================================================== */

#if EXP_MA_T33
/*
 * ~RmXGAt@ MA-T33 ɕt̂
 *
 * "d"     --- L2_12_112_2_2_12_112_12_2_11111112_2_2_2_2_2_2_2_1
 * ""     --- L2_12_112_2_2_12_112_12_2_1111112_12_2_2_2_2_2_12_
 * "^C}[" --- L2_12_112_2_2_12_112_12_2_111112_12_2_2_2_2_2_12_1
 * "U"   --- L2_12_112_2_2_12_112_12_2_1111112_2_2_2_2_2_2_2_11
 * "[h"   --- L2_12_112_2_2_12_112_12_2_111112_112_2_2_2_2_12_2_
 *
 *                                     4.56msec    1.60ms  480us
 * ---+                            +------------+ +------+ +--+ +-
 *    |                            |            | |      | |  | |
 *    |                            |            | |      | |  | |
 *    +----------------------------+            +-+      +-+  +-+
 *               9.00msec                      680usec
 * 
 *    |<========================================>|<======>|<==>|
 *                       "L"                        "2_"    "1"
 * 
 */
const char *F_POWR = "L2_12_112_2_2_12_112_12_2_11111112_2_2_2_2_2_2_2_1";
const char *F_FLOW = "L2_12_112_2_2_12_112_12_2_1111112_12_2_2_2_2_2_12_";
const char *F_TIMR = "L2_12_112_2_2_12_112_12_2_111112_12_2_2_2_2_2_12_1";
const char *F_SWNG = "L2_12_112_2_2_12_112_12_2_1111112_2_2_2_2_2_2_2_11";
const char *F_MODE = "L2_12_112_2_2_12_112_12_2_111112_112_2_2_2_2_12_2_";

/*
 * make first long pulse
 */
void make_long(void)
{
	VOLATILE u_long ttt;

	sct_enable_carrier();
	for(ttt=0; ttt<(myLOOP_WAIT_1msec * 9000 / 1000); ttt++) {
		;
	}
	sct_disable_carrier();

	for(ttt=0; ttt<(myLOOP_WAIT_1msec * 4560 / 1000); ttt++) {
		;
	}

	sct_enable_carrier();
	for(ttt=0; ttt<(myLOOP_WAIT_1msec *  680 / 1000); ttt++) {
		;
	}
	sct_disable_carrier();
}

/*
 * make wide pulse for '2'
 */
void make_p2(void)
{
	VOLATILE u_long ttt;

	for(ttt=0; ttt<(myLOOP_WAIT_1msec * 1600 / 1000); ttt++) {
		;
	}

	sct_enable_carrier();
	for(ttt=0; ttt<(myLOOP_WAIT_1msec *  680 / 1000); ttt++) {
		;
	}
	sct_disable_carrier();
}

/*
 * make narrow pulse for '1'
 */
void make_p1(void)
{
	VOLATILE u_long ttt;

	for(ttt=0; ttt<(myLOOP_WAIT_1msec *  480 / 1000); ttt++) {
		;
	}

	sct_enable_carrier();
	for(ttt=0; ttt<(myLOOP_WAIT_1msec *  680 / 1000); ttt++) {
		;
	}
	sct_disable_carrier();
}

/*
 * a break for interval
 */
void make_break(void)
{
	VOLATILE u_long ttt;

	for(ttt=0; ttt<(myLOOP_WAIT_1msec * 500); ttt++) {
		;
	}
}

/*
 * generate pulses for specified function
 */
void make_function(char const *f)
{
	while(1) {
		switch(*f++) {
		case 'L':
				make_long();
				break;
		case '2':
				make_p2();
				break;
		case '_':
				/* ignore, only skip */
				break;
		case '1':
				make_p1();
				break;
		default:
				return;		/* <=== return */
		}
	}
}

/*
 *
 */
void make_request(char c)
{
	switch(c) {
			/* send indivisual function */
	case '1':
	case 'p':
			make_function(F_POWR);
			break;
	case '2':
	case 'f':
			make_function(F_FLOW);
			break;
	case '3':
	case 't':
			make_function(F_TIMR);
			break;
	case '4':
	case 's':
			make_function(F_SWNG);
			break;
	case '5':
	case 'm':
			make_function(F_MODE);
			break;

			/* send series of functions */
	case 'L':
	case 'S':
	default:
			make_function(F_POWR);
			make_break();
			make_function(F_TIMR);
			make_break();
			make_function(F_SWNG);
			make_break();
			make_function(F_TIMR);
			make_break();
			if(c == 'S') {
				make_function(F_FLOW);
				make_break();
			}
			make_function(F_TIMR);
			make_break();
			break;
	}
}
#endif /* EXP_MA_T33 */

/* =============================================================== */

/*
 *
 */
int main(void)
{
	u_long sysCLK;
	u_long mainCLK;
	int push_counter;
	int sub_counter;
	int i, c;
#if EXP_MA_T33
	static int power_on = 1;
			/*
			 * before installing power-down behaivior,
			 * I had thought to control this by power-switch.
			 */
#endif /* EXP_MA_T33 */

	/*
	 * mask all interrupts by PRIMASK flag in the core
	 */
	__disable_irq();

	/*
	 * initial my variables
	 */
	systick_raw = 0;

	/*
	 * configure I/O pins
	 */
	configure_fixed_pins();

	/*
	 * setup system clock for IRC 12MHz
	 */
#if EXP_MA_T33
	sysCLK = initial_sysCLK(/*use_PLL_or_set_divider=*/ -DIV);
#else /* EXP_MA_T33 */
	sysCLK = initial_sysCLK(/*use_PLL_or_set_divider=*/ 0);
#endif /* EXP_MA_T33 */
	mainCLK = sysCLK * LPC_SYSCTL->SYSAHBCLKDIV;

	/*
	 * configure GPIO after clock fixed
	 */
	configure_GPIO(mainCLK);

	/*
	 * setup uart for debug, etc.
	 */
	cm_init_uart(57600, mainCLK);

	/*
	 *
	 */
	rprintf(
		""							"\r\n"
		"/*"						"\r\n"
		" * hello...(%dMHz/%dMHz)"	"\r\n"
		" */"						"\r\n"
		""							"\r\n"
		, mainCLK / 1000 / 1000
		, sysCLK / 1000 / 1000
	);

	/*
	 * initial peripherals
	 */
#if EXP_MA_T33
	LPC_GPIO_PORT->CLR[0] = GPO_IR;
	LPC_GPIO_PORT->DIR[0] |= GPO_IR;		/* always drive LOW */
#else /* EXP_MA_T33 */
	uart1_c_init(2400, mainCLK);
#endif /* EXP_MA_T33 */
	sctimer_c_init(sysCLK);
	init_SysTick(sysCLK);

#if POWER_DOWN
	/*
	 * setup interrupt on GPI_pSW / GPIO0_1
	 */
	LPC_SYSCTL->PINTSEL[0] = 1;		/* gen PININT0 by GPIO0_1 */
	LPC_PININT->ISEL &= ~(1<<0);		/* edge sensing */
	LPC_PININT->IENF |=  (1<<0);		/* enable falling-edge inteerrupt */
	LPC_PININT->FALL =   (1<<0);		/* clear request */
	NVIC_SetPriority(PININT0_IRQn, PRI_SYSTICK);
	NVIC->ISER[0] = (1 << PININT0_IRQn);	/* enable PININT0 */
	LPC_SYSCTL->STARTERP0 = (1<<0);	/* wakeup by PININT0 */
#endif /* POWER_DOWN */

	/*
	 * start my job
	 */
	push_counter = 0;
	sub_counter = 0;

	__enable_irq();

	while(1) {
#if EXP_MA_T33
		if(power_on) {
			c = 'S';
			power_on = 0;
		} else {
			c = -1;			/* no event to send */
		}
#else /* EXP_MA_T33 */
		c = -1;			/* no event to send */
#endif /* EXP_MA_T33 */

		/*
		 * sense uart arrival
		 */
		i = cm_keyin();
		if(i != -1) {
			c = i;		/* send the char just received */
		}

		/*
		 * sense the switch
		 */
		if(systick_raw) {
			/*
			 * get 10msec trigger
			 */
			systick_raw = 0;

			/*
			 * parse the SW condition
			 *   short push -->      'S'
			 *   long push -->       'L'
			 *   continue pushed --> 'S', 'S', 'S', ...
			 */
#define T_TICK       (1000 / SYSTICK_Hz)			/* [msec] */
#define T_PUSH_SHORT   (  50/*[msec]*/ / T_TICK)		/* [n] */
#define T_PUSH_LONG    ( 500/*[msec]*/ / T_TICK)		/* [n] */
#define T_PUSH_REPEAT  (1500/*[msec]*/ / T_TICK)		/* [n] */
#define T_REPEAT       ( 150/*[msec]*/ / T_TICK)		/* [n] */

			if(LPC_GPIO_PORT->PIN[0] & GPI_pSW) {
				/*
				 * not pushed or released, examine length being pushed
				 */
				if(push_counter >= T_PUSH_REPEAT) {
					;		/* was in repeat mode */
				}
				else if(push_counter >= T_PUSH_LONG) {
					c = 'L';
				}
				else if(push_counter >= T_PUSH_SHORT) {
					c = 'S';
				}
				else {
					;		/* to ignore (push_counter = 0) */
				}

				/*
				 * reset both counter
				 */
				push_counter = 0;
				sub_counter = 0;
			}
			else {
				/*
				 * being pushed
				 */
				push_counter++;
				if(push_counter >= T_PUSH_REPEAT) {
					/*
					 * hold the counter at here,
					 * this also indicate that it entered repeat mode
					 */
					push_counter = T_PUSH_REPEAT;

					/*
					 * count repeat interval
					 */
					sub_counter--;
					if(sub_counter <= 0) {
						c = 'S';
						sub_counter = T_REPEAT;
					}
				}
			}
		}

		/*
		 * send the event if exist
		 */
		if(c != -1) {
#if EXP_MA_T33
			make_request(c);
#else /* EXP_MA_T33 */
			/*
			 * drive the Ir-LED,
			 * modulation is enabled only while sending
			 */
			sct_enable_carrier();
			u1_Putchar(0x55);		/* header */
			u1_Putchar(c);			/* (4.6msec/char) */
			u1_Putchar(0xaa);		/* tailer */
			sct_disable_carrier();
				/*
				 * it takes 15msec to send data
				 * and I will ignore misc input while sending...
				 */
#endif /* EXP_MA_T33 */

			/*
			 * echo back
			 */
			cm_putchar(c);
		}

#if POWER_DOWN
		/*
		 * enter power-down when become idle
		 */
		if((push_counter == 0) && (sub_counter == 0)) {
			LPC_PMU->PCON = 0x02;		/* select power-down mode */
			SCB->SCR = 0x04;			/*    use deep sleep */
			__WFI();					/* Wair_For_Interrupt */
			systick_raw = 1;
		}
#endif /* POWER_DOWN */
	}

	/* never return */
}

/* =============================================================== */

/*
 * determine how handle fixed pins
 */
void configure_fixed_pins(void)
{
	u_long tmp;

	tmp = LPC_SWM->PINENABLE0;

	/*
	 * disable these fixed-pin-function, GPIO will have control
	 */
	tmp |= 0
			| (1<<2)		/* disable SWCLK, PIO0_3 enabled */
			| (1<<3)		/* disable SWDIO, PIO0_2 enabled */
		//  | (1<<6)		/* disable RESET, PIO0_5 enabled */
			| 0;

	/*
	 * enable these fixed-pin-function, GPIO will lose control
	 *   (do not enable both ACMP_I2 and CLKIN at same time)
	 */
	tmp &= ~(0
		//  | (1<<0)		/* enable ACMP_I1, PIO0_0 disabled */
		//  | (1<<1)		/* enable ACMP_I2, PIO0_1 disabled */
//NO-PIN//  | (1<<4)		/* enable XTALIN,  PIO0_8 disabled */
//NO-PIN//  | (1<<5)		/* enable XTALOUT, PIO0_9 disabled */
#if USE_CLKIN_12MHz
			| (1<<7)		/* enable CLKIN,   PIO0_1 disabled */
#else /* USE_CLKIN_12MHz */
		//  | (1<<7)		/* enable CLKIN,   PIO0_1 disabled */
#endif /* USE_CLKIN_12MHz */
//NO-PIN//  | (1<<8)		/* enable VDDCMP,  PIO0_6 disabled */
			| 0);

	/*
	 * write back
	 */
	LPC_SWM->PINENABLE0 = tmp;
}

/* =============================================================== */

/*
 * configure system clock
 *   (param > 0) start PLL to run at 30MHz
 *   (param ==0) run at IRC 12MHz
 *   (param < 0) run at IRC 12MHz with divider
 */
	/*
	 * set_pll() ROM-API can not configure 30MHz from 12MHz...
	 *   (he can configure only 1.5x, 2x, 3x of input frequency)
	 */
u_long initial_sysCLK(int use_PLL_or_set_divider)
{
	u_long divider;

	/*
	 * expect these conditions by cold start
	 */
//	LPC_SYSCTL->SYSAHBCLKDIV = 1;		/* 1/1 (reset value) */
//	LPC_SYSCTL->SYSPLLCLKSEL = 0;		/* sel IRC (reset value) */
//	LPC_SYSCTL->SYSPLLCLKUEN = 0;
//	LPC_SYSCTL->SYSPLLCLKUEN = 1;
//	LPC_SYSCTL->MAINCLKSEL = 0x0;		/* run with IRC (reset value) */
//	LPC_SYSCTL->MAINCLKUEN = 0;
//	LPC_SYSCTL->MAINCLKUEN = 1;

#if USE_CLKIN_12MHz
	/*
	 * use external 12MHz thru CLKIN
	 */
	LPC_SYSCTL->SYSPLLCLKSEL = 0x3;		/* CLKIN */
	LPC_SYSCTL->SYSPLLCLKUEN = 0;
	LPC_SYSCTL->SYSPLLCLKUEN = 1;
	LPC_SYSCTL->MAINCLKSEL = 0x1;		/* PLL clock-in */
	LPC_SYSCTL->MAINCLKUEN = 0;
	LPC_SYSCTL->MAINCLKUEN = 1;
#endif /* USE_CLKIN_12MHz */

	if(use_PLL_or_set_divider > 0) {
		/*
		 * start PLL for mainCLK=60MHz (12MHz * 5 = 240MHz / (2 * 2))
		 */
		LPC_SYSCTL->SYSPLLCTRL = (/*PSEL=*/ 0x01 <<5) | (/*MSEL=*/ 5 -1);
		LPC_SYSCTL->PDRUNCFG &= ~(1<<7);		/* enable PLL */
		while(! (LPC_SYSCTL->SYSPLLSTAT & 1) ) {
			;	/* wait for lock */
		}

		/*
		 * connect PLL output
		 */
		LPC_SYSCTL->SYSAHBCLKDIV = 2;		/* 1/2 */
		LPC_SYSCTL->MAINCLKSEL = 0x3;		/* select PLL-output */
		LPC_SYSCTL->MAINCLKUEN = 0;
		LPC_SYSCTL->MAINCLKUEN = 1;

		return(30*1000*1000);	/* sysCLK is: 30MHz */
	}
	else {
		divider = - use_PLL_or_set_divider;
		if(divider == 0) {
			divider = 1;
		} else if(divider > 255) {
			divider = 255;
		}
		LPC_SYSCTL->SYSAHBCLKDIV = divider;
		return(12*1000*1000 / divider);		/* return sysCLK */
	}
}

/* =============================================================== */

/*
 *
 */
void configure_GPIO(u_long mainCLK)
{
	u_long div;
	u_long n, tmp;

	/*
	 * prevent the the open-drain pins from internally floating
	 *   (PIO0_10 and PIO0_11, section 6.3, UM10601)
	 */
#define INTERNAL_OPEN_DRAIN_PINS   ((1<<10) | (1<<11))
	LPC_GPIO_PORT->CLR[0]  = INTERNAL_OPEN_DRAIN_PINS;	/* both Low */
	LPC_GPIO_PORT->DIR[0] |= INTERNAL_OPEN_DRAIN_PINS;	/* start to drive */

	/*
	 * setup GPIO
	 */
	LPC_GPIO_PORT->DIR[0] &= ~GPI_pSW;		/* confirm input */

	/*
	 * use glitch filter for the input
	 */
	/*
	 * setup filter clock: maybe 12MHz / 240 = 50kHz
	 */
	div = mainCLK / 50 * 1000;			/* 50kHz */
	if(div > 255) {
		div = 255;
	}
	LPC_SYSCTL->IOCONCLKDIV[6 - 0] = div;		/* select CLK0 */

	/*
	 * setup I/O configuration
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_IOCON);

	n = IOCON_PIO1;
	tmp = LPC_IOCON->PIO0[n];
	tmp |= PIN_HYS_MASK;						/* add hysteresis */
#if POWER_DOWN
		/* clock is stopped while power-down */
#else /* POWER_DOWN */
	tmp &= ~PIN_SMODE_MASK;
	tmp |= (PIN_SMODE_CYC3 <<PIN_SMODE_BITNUM);		/* filter: 3clock cycle */
	tmp &= ~PIN_CLKDIV_MASK;
	tmp |= (IOCONCLKDIV0 << PIN_CLKDIV_BITNUM);		/* select CLK0 */
#endif /* POWER_DOWN */
	LPC_IOCON->PIO0[n] = tmp;

	LPC_SYSCTL->SYSAHBCLKCTRL &= ~(1 << SYSCTL_CLOCK_IOCON);
}

/* =============================================================== */

/*
 * run SysTick timer
 */
void init_SysTick(u_long sysCLK)
{
	/*
	 * setup SysTick timer
	 */
	SysTick->LOAD = sysCLK / SYSTICK_Hz / 2 - 1;	/* use half-clock */
	SysTick->VAL = 0;			/* clear the counter (value is ignored) */
		/*
		 * 30,000,000 = 0x01C9_C380  >  (1<<24)
		 * 16,777,215 = 0x00FF_FFFF  =  (1<<24)
		 * 15,000,000 = 0x00E4_E1C0  <  (1<<24) ... half-clock covers 1Hz
		 */
	/*
	 * setup NVIC
	 */
	NVIC_SetPriority(SysTick_IRQn, PRI_SYSTICK);

	/*
	 * enable intr and run it
	 */
	SysTick->CTRL = 0x03;		/* run it with half-clock, intr enabled */
}

/*
 * interrupt service routine
 */
void SysTick_Handler(void)
{
	systick_raw = 1;
		/*
		 * interrupt request flag may be cleared automatically
		 */
}

#if POWER_DOWN
/*
 * interrupt service routine
 */
void PININT0_Handler(void)
{
	LPC_PININT->FALL = (1<<0);		/* clear request */
}
#endif /* POWER_DOWN */

/* =============================================================== */

/*
 * try to catch exceptions
 */

void def_Exception_Handler(void)
{
	rprintf("! some Exception\r\n");
	while(1)
		;
	/* never return */
}

void def_NMI_Handler(void)
{
	rprintf("! NMI\r\n");
	while(1)
		;
	/* never return */
}

void def_HardFault_Handler(void)
{
	rprintf("! HardFault\r\n");
	while(1)
		;
	/* never return */
}

void def_SVCall_Handler(void)
{
	rprintf("! SVCall\r\n");
	while(1)
		;
	/* never return */
}

void def_PendSV_Handler(void)
{
	rprintf("! PendSV\r\n");
	while(1)
		;
	/* never return */
}

/* =============================================================== */
